"""
Class to contain a text sorting rule.
SortRule = namedtuple('SortRule', ['folder', 'regex'])
"""
import csv
import importlib.resources
import re
import sys
from dataclasses import dataclass
from os import environ
from pathlib import Path
from typing import List, Optional, Union

from rich.console import Console
from rich.text import Text

from clown_sort.util.constants import CRYPTO, PACKAGE_NAME
from clown_sort.util.rich_helper import print_error

RULES_CSV_PATHS = 'RULES_CSV_PATHS'
SORTING_RULES_DIR = importlib.resources.files(PACKAGE_NAME).joinpath('sorting_rules')
CRYPTO_RULES_CSV_PATH = Path(str(SORTING_RULES_DIR.joinpath('crypto.csv')))


class SortRuleParseError(RuntimeError):
    pass


@dataclass
class SortRule:
    folder: str
    regex: Union[str, re.Pattern]

    def __post_init__(self):
        try:
            self.regex = re.compile(self.regex, re.IGNORECASE | re.MULTILINE)
        except re.error as e:
            msg = f"{str(e)} while processing '{self.folder}' sort rule '{self.regex}'"
            raise SortRuleParseError(msg) from e

    @classmethod
    def load_rules_csv(cls, file_path: Path) -> List['SortRule']:
        """Turn a CSV of sort rules into a list of SortRule objects."""
        sort_rules = []

        with open(Path(file_path), mode='r') as csvfile:
            for row in csv.DictReader( filter(SortRule.is_valid_row, csvfile), delimiter=','):
                try:
                    sort_rules.append(cls(row['folder'], row['regex']))
                except SortRuleParseError as e:
                    print_error(f"{str(e)} in file '{file_path}'!")
                    raise e

        return sort_rules

    @staticmethod
    def is_valid_row(row: str) -> bool:
        """Returns false for whitespace rows and those that start with '#'."""
        stripped_row = row.strip()
        return len(stripped_row) > 0 and stripped_row[0] != '#'

    @staticmethod
    def sort_rules_csvs(csv_paths: Optional[str]) -> List[Path]:
        """Expects str of paths separated by ':'. Replace the string 'crypto' with the default CSV."""
        if isinstance(csv_paths, Path):
            return [csv_paths]
        elif csv_paths is None:
            return SortRule.default_rules_csv_paths()

        return [
            CRYPTO_RULES_CSV_PATH if csv_path == CRYPTO else Path(csv_path)
            for csv_path in csv_paths.split(':')
        ]

    @staticmethod
    def default_rules_csv_paths() -> List[Path]:
        """Check for RULES_CSV_PATHS in env vars otherwise return default crypto rules."""
        return SortRule.sort_rules_csvs(environ.get(RULES_CSV_PATHS, CRYPTO_RULES_CSV_PATH))



"""
Global configuration.
"""
import re
import sys
from argparse import Namespace
from importlib.metadata import version
from os import environ
from pathlib import Path
from typing import List, Optional, Union

from rich import box
from rich.console import Console
from rich.table import Table
from rich.text import Text

from clown_sort.sort_rule import SortRule, SortRuleParseError
from clown_sort.util import rich_helper
from clown_sort.util.argument_parser import parser
from clown_sort.util.constants import PACKAGE_NAME
from clown_sort.util.filesystem_helper import create_dir_if_it_does_not_exist, subdirs_of_dir
from clown_sort.util.logging import log, set_log_level


class Config:
    debug: bool = False
    dry_run: bool = True
    manual_sort: bool = False
    manual_fallback: bool = False
    only_if_match: bool = False
    leave_in_place: bool = False
    screenshots_only: bool = True
    delete_originals: bool = False
    rescan_sorted: bool = False
    yes_overwrite: bool = False
    print_as_parsed: bool = False
    sort_rules: List[SortRule] = []
    filename_regex: re.Pattern

    @classmethod
    def configure(cls):
        """Parse arguments and configure."""
        if '--version' in sys.argv:
            print(f"{PACKAGE_NAME} {version(PACKAGE_NAME)}")
            sys.exit()

        args: Namespace = parser.parse_args()

        if args.debug:
            cls.enable_debug_mode()

        Config.filename_regex = re.compile(args.filename_regex)
        Config.leave_in_place = True if args.leave_in_place else False
        Config.only_if_match = True if args.only_if_match else False
        Config.delete_originals = True if args.delete_originals else False
        Config.rescan_sorted = True if args.rescan_sorted else False
        Config.yes_overwrite = True if args.yes_overwrite else False

        screenshots_dir = Path(args.screenshots_dir).expanduser()
        destination_dir = Path(args.destination_dir or args.screenshots_dir).expanduser()
        rules_csvs = SortRule.sort_rules_csvs(args.rules_csv)
        log.debug(f"Rules CSVs: {rules_csvs}")
        Config.set_directories(screenshots_dir, destination_dir, rules_csvs)

        if Config.leave_in_place and Config.delete_originals:
            Console().print("--leave-in-place and --delete-originals are mutually exclusive.", style='red')
            sys.exit(-1)

        if args.show_rules:
            Console().print(cls._rules_table())
            sys.exit()

        if args.execute:
            Config.dry_run = False
            rich_helper.is_dry_run = False
        else:
            print("Dry run...")

        if args.all:
            print("Processing all files in directory, not just 'Screenshot' files....")
            Config.screenshots_only = False

        if args.manual_sort or args.manual_fallback:
            _check_for_pysimplegui()

            if args.manual_sort:
                Config.manual_sort = True
            if args.manual_fallback:
                if Config.only_if_match:
                    Console().print('Only one of --manual-fallback and --only-if-match can be specified.', style='red')
                    sys.exit(-1)

                Config.manual_fallback = True

    @classmethod
    def set_directories(
            cls,
            screenshots_dir: Path,
            destination_dir: Path,
            rules_csv_paths: List[Path]
    ) -> None:
        """Set the directories to find screenshots in and sort screenshots to."""
        screenshots_dir = Path(screenshots_dir)
        destination_dir = Path(destination_dir or screenshots_dir)
        rules_csv_paths = [Path(r) for r in rules_csv_paths]

        for csv_path in rules_csv_paths:
            if not csv_path.is_file():
                print(f"ERROR: '{csv_path}' is not a file.")
                sys.exit(-1)

            try:
                cls.sort_rules += SortRule.load_rules_csv(csv_path)
            except SortRuleParseError:
                sys.exit(-1)

        cls.screenshots_dir: Path = Path(screenshots_dir)
        cls.destination_dir: Path = Path(destination_dir or screenshots_dir)
        cls.sorted_screenshots_dir = cls.destination_dir.joinpath('Sorted')
        cls.processed_screenshots_dir = cls.destination_dir.joinpath('Processed')
        cls.pdf_errors_dir = cls.destination_dir.joinpath('PDF Errors')

        for dir in [cls.destination_dir, cls.sorted_screenshots_dir, cls.processed_screenshots_dir]:
            create_dir_if_it_does_not_exist(dir)

        cls._log_configured_paths()

    @classmethod
    def get_sort_dirs(cls) -> List[str]:
        """Returns a list of the subdirectories already created for sorted images."""
        return sorted(subdirs_of_dir(cls.sorted_screenshots_dir), key=lambda d: d.lower())

    @classmethod
    def enable_debug_mode(cls) -> None:
        Config.debug = True
        set_log_level('DEBUG')

    @classmethod
    def _rules_table(cls) -> Table:
        """Generate a table of the sort rules in effect."""
        table = Table(
            'Folder', 'Regex',
            title='Sorting Rules',
            title_style='color(153) italic dim',
            header_style='color(245)',
            box=box.SIMPLE,
            show_edge=False,
            collapse_padding=True
        )

        for sort_rule in Config.sort_rules:
            table.add_row(sort_rule.folder, sort_rule.regex.pattern)

        table.columns[0].style = 'bright_red'
        table.columns[1].style = 'color(65)'
        return table

    @classmethod
    def _log_configured_paths(cls) -> None:
        log.debug(f"screenshots_dir: {cls.screenshots_dir}")
        log.debug(f"destination_dir: {cls.destination_dir}")
        log.debug(f"sorted_screenshots_dir: {cls.sorted_screenshots_dir}")
        log.debug(f"processed_screenshots_dir: {cls.processed_screenshots_dir}")
        log.debug(f"pdf_errors_dir: {cls.pdf_errors_dir}")


def _check_for_pysimplegui():
    try:
        import PySimpleGUI as sg
    except ModuleNotFoundError:
        msg = Text('ERROR', style='bright_red')

        msg.append(
            ': PySimpleGUI package must be installed before you can use the manual selector. Try running:',
            style='bright_white'
        )

        log_optional_module_warning('gui', msg)
        console = Console(color_system='256')
        #console.line()
        console.print(f"You make also need to install 'python-tk'. In macOS this can be installed with 'brew install python-tk'.")
        sys.exit()


def check_for_pymupdf() -> bool:
    try:
        import fitz
        return True
    except ModuleNotFoundError:
        log_optional_module_warning('pdf')
        return False


# TODO: it sucks that this is here but it's dependency hell otherwise
def log_optional_module_warning(module_name: str, msg: Optional[Text] = None) -> None:
    """msg is optional argument for a custom message, otherwise it's a warning"""
    if msg is None:
        msg = Text('WARNING', style='bright_red').append(
            f": Optional package '{module_name}' not installed. Try running:",
            style='bright_white'
        )

    console = Console(color_system='256')
    console.line()
    console.print(msg)
    console.line()
    console.print(f"     pipx install clown_sort\\[{module_name}] --force", style='bright_cyan')
    console.line()
    console.print(f"'pip install' works if you're not using pipx. Use 'poetry install --all-extras' if you're in a development environment.")





"""
Entry point for all of the clown_sort scripts.
"""
from argparse import Namespace
from glob import glob
from os import environ, getcwd, path
from pathlib import Path
from typing import List, Union

from dotenv import load_dotenv

# load_dotenv() should be called as soon as possible (before parsing local classes) but not for pytest
if not environ.get('INVOKED_BY_PYTEST', False):
    for dotenv_file in [path.join(dir, '.clown_sort') for dir in [getcwd(), Path.home()]]:
        if path.exists(dotenv_file):
            load_dotenv(dotenv_path=dotenv_file)
            break

from clown_sort.util.argument_parser import parse_text_extraction_args, parse_pdf_page_extraction_args
from clown_sort.config import Config
from clown_sort.files.image_file import ImageFile
from clown_sort.files.pdf_file import PdfFile
from clown_sort.files.sortable_file import SortableFile
from clown_sort.sort_selector import process_file_with_popup
from clown_sort.util.filesystem_helper import (IMAGE_FILE_EXTENSIONS, files_in_dir, is_image,
      is_pdf, set_timestamp_based_on_screenshot_filename)
from clown_sort.util.logging import log
from clown_sort.util.rich_helper import console


def sort_screenshots():
    Config.configure()

    if Config.rescan_sorted:
        _rescan_sorted_screenshots()
        return

    for file_to_sort in screenshot_paths(Config.screenshots_dir):
        if Config.manual_sort:
            if file_to_sort._can_be_presented_in_popup():
                process_file_with_popup(file_to_sort)
            else:
                print(f"'{file_to_sort.file_path}' is not suitable for manual sort, skipping...")
        else:
            file_to_sort.sort_file()


def extract_text_from_files() -> None:
    """
    Extract text from a single file or from all files in a given directory. Can accept
    multiple paths as arguments on the command line.
    """
    args: Namespace = parse_text_extraction_args()
    console.line()

    if args.debug:
        Config.enable_debug_mode()
    if args.print_as_parsed:
        Config.print_as_parsed = True

    for file_path in args.files_to_process:
        sortable_file = build_sortable_file(file_path)

        if isinstance(sortable_file, PdfFile):
            sortable_file.print_extracted_text(page_range=args.page_range)
        else:
            sortable_file.print_extracted_text()

        console.line(2)


def extract_pages_from_pdf() -> None:
    args = parse_pdf_page_extraction_args()
    PdfFile(args.pdf_file).extract_page_range(args.page_range, destination_dir=args.destination_dir)


def set_screenshot_timestamps_from_filenames():
    """Parse the filenames to reset the file creation timestamps."""
    Config.configure()

    for image in screenshot_paths(Config.screenshots_dir):
        set_timestamp_based_on_screenshot_filename(image.file_path)


def _rescan_sorted_screenshots():
    """Rescan sorted folders."""
    console.print(f"Rescanning '{Config.sorted_screenshots_dir}'...")
    sortable_files: List[SortableFile] = []
    file_paths: List[str] = []

    for extname in IMAGE_FILE_EXTENSIONS:
        for pattern in ['*', '**/*']:
            glob_pattern = Config.sorted_screenshots_dir.joinpath(f"{pattern}{extname}")
            log.debug(f"Adding '{glob_pattern}' to glob patterns...")
            file_paths.extend(glob(str(glob_pattern)))

    for file_path in file_paths:
        if Config.screenshots_only and not Config.filename_regex.match(path.basename(file_path)):
            log.debug(f"Skipping '{file_path}' because it doesn't match the filename_regex...")
            continue

        sortable_files.append(build_sortable_file(file_path))

    console.print(
        f"Re-processing {len(sortable_files)} files in '{Config.sorted_screenshots_dir}'...",
        style='bright_green'
    )

    for file_path in sortable_files:
        file_path.sort_file()


def screenshot_paths(dir: Path) -> List[SortableFile]:
    """Returns a list of ImageFiles for all the screenshots to be sorted."""
    screenshots = [
        build_sortable_file(f) for f in files_in_dir(Config.screenshots_dir)
        if not Config.screenshots_only or Config.filename_regex.match(path.basename(f))
    ]

    return sorted(screenshots, key=lambda f: f.basename)


def build_sortable_file(file_path: Union[str, Path]) -> SortableFile:
    """Decide if it's a PDF, image, or other type of file."""
    if is_image(file_path):
        return ImageFile(file_path)
    elif is_pdf(file_path):
        return PdfFile(file_path)
    else:
        return SortableFile(file_path)



"""
Decide on a filename string based on some OCR text.
"""
import re
from difflib import SequenceMatcher
from typing import Optional

from clown_sort.util.logging import log
from clown_sort.util.filesystem_helper import strip_bad_chars, strip_mac_screenshot

TWEET_REPLY_REGEX = re.compile(
    'Replying to (@\\w{3,15}).*?\\n(?P<body>.*)',
    re.DOTALL | re.MULTILINE
)

TWEET_REGEX = re.compile(
    '(@\\w{3,15}(\\.\\.\\.)?)(\\s{1,2}-\\s{1,2}([\\dti]{1,2}[smhd]|Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec).*?)?\\n(?P<body>.*)',
    re.DOTALL | re.MULTILINE
)

REDDIT_POST_REGEX = re.compile(
    '(r/(?P<sub>\\w{3,30}) - )?Posted by u/(?P<author>\\w{3,30}) \\d+.*?\\n(?P<body>.*)',
    re.DOTALL | re.MULTILINE
)

REDDIT_REPLY_REGEX = re.compile(
    '(?P<author>\\w{3,30})( OP -)?\\s+(.*?)-\\s+\\d+\\s+(seconds|min(\\.|utes)|hours|hr\\.|days|months|years)\\s+ago\\s*(.*?)\\n(?P<body>.*?)(Reply\\s+)?Give\\s?Award\\s+Share\\s+Report',
    re.DOTALL | re.MULTILINE | re.IGNORECASE
)

REVEDDIT_REGEX = re.compile('Reveddit Real.?Time')
SUBREDDIT_REGEX = re.compile('/r/(?P<subreddit>\\w+)|\\sse(lf|ir)\\.(?P<subreddit2>\\w+)')
DUNE_ANALYTICS_REGEX = re.compile('Query results (.*) @\\w+')
RETWEETED_REGEX = re.compile('(.*) Retweeted')

MAX_FILENAME_LENGTH = 225
MIN_LENGTH_FOR_DUPE_CHECK = 9
DEFAULT_LENGTH_FOR_LONG_FILENAMES = 190
MIN_SIMILARITY_RATIO_TO_BE_SAME = 0.80


class FilenameExtractor:
    def __init__(self, image_file: 'ImageFile') -> None:
        self.image_file = image_file
        self.text: Optional[str] = image_file.extracted_text()
        self.basename_length: int = len(image_file.basename)
        self.available_char_count: int = MAX_FILENAME_LENGTH - self.basename_length - 1
        self.author: Optional[str] = None
        self.reply_to_account: Optional[str] = None

    def filename(self) -> str:
        """Examine self.text and decide on an appropriate filename."""
        if self.text is None:
            return self.image_file.basename

        if DUNE_ANALYTICS_REGEX.search(self.text):
            dune_match = DUNE_ANALYTICS_REGEX.search(self.text)
            query_title = strip_bad_chars(dune_match.group(1))
            filename_str = 'Dune Analytics "' + query_title + '" '
            new_filename = filename_str + self.image_file.basename
        elif self._is_tweet() or self._is_reddit():
            if self._is_tweet():
                filename_str = self._filename_str_for_tweet()
            else:
                filename_str = self._filename_str_for_reddit()

            filename_str = filename_str[0:-1] if filename_str.endswith('.') else filename_str
            filename = f"{filename_str} {self.image_file.basename_without_ext}"
            new_filename = filename + self.image_file.extname
        elif self._is_reveddit():
            filename_str = 'Reveddit '
            subreddit_match = SUBREDDIT_REGEX.search(self.text)

            if subreddit_match is not None:
                filename_str += 'r_' + (subreddit_match.group('subreddit') or subreddit_match.group('subreddit2'))

            new_filename = filename_str[0:self.available_char_count]  + ' ' + self.image_file.basename
        else:
            filename_str = self.text[0:max(100, self.available_char_count)]
            new_filename = self._build_filename(self.image_file.basename_without_ext, self.text)
            new_filename = new_filename[0:MAX_FILENAME_LENGTH] + self.image_file.extname

        if self._is_text_already_in_filename(filename_str):
            return self.image_file.basename

        if len(new_filename) > 255:
            raise ValueError(f"'{new_filename}' is {len(new_filename)} chars")

        return new_filename

    def _is_tweet(self) -> bool:
        """Return true if the text looks like a tweet."""
        # TODO: the check for @crypto_oracle is a hack
        return TWEET_REGEX.search(self.text) is not None and '@crypto_oracle' not in self.text

    def _is_retweet(self) -> bool:
        """Return true if the text looks like a retweet."""
        return self._is_tweet() and RETWEETED_REGEX.search(self._first_line()) is not None

    def _retweeter(self) -> Optional[str]:
        """Return the name of the retweeter (or None if there isn't one)"""
        if self._is_retweet():
            retweet_match = RETWEETED_REGEX.search(self._first_line())
            retweeter = retweet_match.group(1)
            # Remove the leading 'tl' (or whatever Tesseract interprets the loop arrow icon to be)
            return ','.join(retweeter.split()[1:])
        else:
            return None

    def _is_reddit(self) -> bool:
        """Return true if it's a reddit post or reddit comment."""
        return self._is_reddit_post() or self._is_reddit_comment()

    def _is_reddit_post(self) -> bool:
        """Return true if it's a reddit post."""
        return REDDIT_POST_REGEX.search(self.text) is not None

    def _is_reddit_comment(self) -> bool:
        """Return true if it's a reddit comment."""
        return REDDIT_REPLY_REGEX.search(self.text) is not None

    def _is_reveddit(self) -> bool:
        """Return true if text is from reveddit.com (site for deleted reddit posts and comments)."""
        return REVEDDIT_REGEX.search(self.text) is not None

    def _filename_str_for_tweet(self) -> str:
        """Build a filename for tweets."""
        tweet_match = TWEET_REGEX.search(self.text)
        self.author = tweet_match.group(1)
        body = tweet_match.group('body')
        filename_text = f"Tweet by {self.author}"
        retweeter = self._retweeter()
        reply_to = TWEET_REPLY_REGEX.search(self.text)

        if retweeter is not None:
            filename_text = f'Retweeted by {retweeter} - {filename_text}'

        if reply_to is not None:
            self.reply_to_account = reply_to.group(1)

            if self.reply_to_account != self.author:
                filename_text += f" replying to {self.reply_to_account}"

            body = reply_to.group('body')

        return self._build_filename(filename_text, body)

    def _filename_str_for_reddit(self) -> str:
        """Build a filename for Reddit posts and comments."""
        if self._is_reddit_post():
            reddit_match = REDDIT_POST_REGEX.search(self.text)
            self.author = reddit_match.group('author')
            subreddit: str = reddit_match.group('sub')
            body: str = reddit_match.group('body')
            filename_text: str = f"Reddit post by {self.author}"

            if subreddit:
                filename_text += f" in {subreddit}"
        else:
            reddit_match = REDDIT_REPLY_REGEX.search(self.text)
            self.author = reddit_match.group('author')
            body: str = reddit_match.group('body')
            filename_text: str = f"Reddit post by {self.author}"

        return self._build_filename(filename_text, body)

    def _build_filename(self, filename_text: str, body: str) -> str:
        """Construct a workable filename."""
        body = strip_bad_chars(body)
        end_position = self.available_char_count - len(filename_text) - 2

        if end_position < 0:
            end_position = DEFAULT_LENGTH_FOR_LONG_FILENAMES

        body = body[0:end_position].strip()

        if len(body) == 0:
            return filename_text

        return f'{filename_text} - "{body}"'.replace('  ', ' ')

    def _is_text_already_in_filename(self, filename_str: str) -> bool:
        """Check if the extracted text is already in the filename"""

        # Subtract 25 to account for the 'Screenshot at 2020-05-05' etc part of filename
        chars_to_compare = min(len(filename_str), len(self.image_file.basename) - 25)
        # In case it's < 0 or v. small
        chars_to_compare = max(chars_to_compare, MIN_LENGTH_FOR_DUPE_CHECK)

        # Cleanup strings before comparing
        clean_filename_str = strip_mac_screenshot(strip_bad_chars(filename_str))[0:chars_to_compare]
        clean_basename = strip_mac_screenshot(self.image_file.basename)[0:len(clean_filename_str)]
        matcher = SequenceMatcher(None, clean_filename_str, clean_basename)
        similarity = matcher.ratio()

        is_duplicate_text = (matcher.ratio() > MIN_SIMILARITY_RATIO_TO_BE_SAME) \
                         or (clean_filename_str in self.image_file.basename)

        msg = f"Similarity of '{clean_filename_str}'\n  and '{clean_basename}'\n" \
              f"  is {similarity} (is_duplicate_text: {is_duplicate_text})"
        log.debug(msg)

        if is_duplicate_text and len(filename_str) > MIN_LENGTH_FOR_DUPE_CHECK:
            log.debug(f"'{clean_filename_str}' already appears in filename, not renaming.")
            log.debug(f"Extracted text: '{self.text}'")
            return True
        else:
            log.debug(f"\n'{clean_filename_str}'\nis not in\n'{self.image_file.basename}'\n")
            return False

    def _first_line(self) -> str:
        return self.text.split('\n')[0]



"""
Methods to help with string operations.
"""
import re


def is_empty(text: str) -> bool:
    return text is None or len(text) == 0


def spaces_to_underscores(text: str) -> str:
    return text.replace(' ', '_')


def exception_str(e: Exception) -> str:
    """A string with the type and message."""
    return f"{type(e).__name__}: {e}"
